import pygad
import random
import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
import time
import os


class CVRPTest:
    def __init__(self, problem_instance):
        self.problem_instance = problem_instance
        self.num_customers = len(problem_instance.node_coord_section)
        self.num_vehicles = problem_instance.num_vehicles
        self.vehicle_capacity = float(problem_instance.capacity)
        self.distance_matrix = self.create_distance_matrix(
            problem_instance.node_coord_section)
        self.demand_dict = {node: demand for node,
                            demand in problem_instance.demand_section}
        self.depot_index = int(problem_instance.depot_index)
        self.best_solution_cost = None
        self.best_fitness_over_generations = []
        self.best_distances_over_generations = []

    @staticmethod
    def euclidean_distance(point1, point2):
        return np.sqrt((point1[0] - point2[0])**2 + (point1[1] - point2[1])**2)

    def are_all_customers_visited(self, solution):
        visited_customers = set(solution)
        return len(visited_customers) == self.num_customers

    def on_generation(self, ga_instance):
        generation = ga_instance.generations_completed
        best_solution_fitness = ga_instance.best_solution()[1]
        best_solution = ga_instance.best_solution()[0]
        self.best_fitness_over_generations.append(best_solution_fitness)
        best_routes = self.generate_routes(best_solution)
        total_distance = self.calculate_total_distance(best_routes)
        self.best_distances_over_generations.append(total_distance)
        print(
            f"Generation {generation}: Best Fitness = {best_solution_fitness}")

    def plot_fitness_over_generations(self):
        plt.plot(self.best_fitness_over_generations)
        plt.title('Best Fitness Value Over Generations')
        plt.xlabel('Generation')
        plt.ylabel('Fitness Value')
        plt.show()

    def plot_distance_over_generations(self):
        plt.plot(self.best_distances_over_generations)
        plt.title('Total Distance of Best Solution Over Generations')
        plt.xlabel('Generation')
        plt.ylabel('Total Distance')
        plt.savefig('./images/CVRP.png')

    def create_distance_matrix(self, node_coord_section):
        num_nodes = len(node_coord_section)
        distance_matrix = np.zeros((num_nodes, num_nodes))

        for i in range(num_nodes):
            for j in range(num_nodes):
                if i != j:
                    distance_matrix[i][j] = self.euclidean_distance(
                        node_coord_section[i][1:], node_coord_section[j][1:])
        return distance_matrix

    def create_initial_population(self):
        initial_population = []
        for _ in range(200):
            # Each chromosome is a list of customer nodes. The GA will need to find the optimal ordering.
            chromosome = np.random.permutation(self.num_customers).tolist()
            initial_population.append(chromosome)
        return initial_population

    # def nearest_neighbor_heuristic(self, start_customer):
    #     # 0-based indexing for customers
    #     unvisited_customers = set(range(self.num_customers))
    #     unvisited_customers.remove(start_customer)
    #     route = [start_customer]
    #     current_customer = start_customer

    #     while unvisited_customers:
    #         next_customer = min(
    #             unvisited_customers,
    #             key=lambda customer: self.distance_matrix[current_customer][customer]
    #         )
    #         route.append(next_customer)
    #         current_customer = next_customer
    #         unvisited_customers.remove(next_customer)

    #     return route

    # def create_initial_population(self):
    #     initial_population = []
    #     for _ in range(200):
    #         # Start with a random customer for the nearest neighbor heuristic
    #         start_customer = np.random.choice(range(self.num_customers))
    #         chromosome = self.nearest_neighbor_heuristic(start_customer)
    #         # Shuffle the rest of the customers and append to the chromosome
    #         remaining_customers = list(
    #             set(range(self.num_customers)) - set(chromosome))
    #         np.random.shuffle(remaining_customers)
    #         chromosome.extend(remaining_customers)
    #         initial_population.append(chromosome)
    #     return initial_population

    def generate_routes(self, solution):
        routes = []
        current_route = []
        vehicle_load = 0

        for customer_index in solution:
            customer = customer_index + 1  # customer indices are offset by 1
            demand = self.demand_dict[customer]
            if vehicle_load + demand > self.vehicle_capacity:
                routes.append(current_route)
                current_route = []
                vehicle_load = 0
            current_route.append(customer)
            vehicle_load += demand
        routes.append(current_route)  # Append the last route
        return routes

    def calculate_total_distance(self, routes):
        total_distance = 0
        visited_customers = set()

        for route in routes:
            for i in range(len(route) - 1):
                total_distance += self.distance_matrix[int(route[i]) -
                                                       1][int(route[i + 1]) - 1]
                visited_customers.add(int(route[i]))
                visited_customers.add(int(route[i + 1]))

            # Add distance from and to the depot
            total_distance += self.distance_matrix[self.depot_index - 1][int(
                route[0]) - 1]
            total_distance += self.distance_matrix[int(route[-1]) -
                                                   1][self.depot_index - 1]

        # Check if all customers are visited
        if len(visited_customers) != self.num_customers:
            missing_customers_count = self.num_customers - \
                len(visited_customers)
            # Assign a high penalty for each missing customer
            # Adjust the penalty value as needed
            penalty = missing_customers_count * 1000
            total_distance += penalty

        return total_distance

    def calculate_average_total_distance(self, population):
        total_distances = [self.calculate_total_distance(
            self.generate_routes(chromosome)) for chromosome in population]
        average_distance = sum(total_distances) / len(total_distances)
        return average_distance

    def fitness_func(self, ga_instance, solution, solution_idx):
        routes = self.generate_routes(solution)
        total_distance = self.calculate_total_distance(routes)
        # Add 1 to avoid division by zero
        fitness = 1.0 / (total_distance + 1.0)
        return fitness

    def get_random_slice_indices(self, array_length):
        slice_index1 = random.randint(0, array_length - 1)
        slice_index2 = random.randint(0, array_length - 1)

        # Ensure slice_index1 is less than slice_index2
        if slice_index1 > slice_index2:
            slice_index1, slice_index2 = slice_index2, slice_index1

        # Optionally, ensure that the slice is not empty
        if slice_index1 == slice_index2:
            if slice_index2 < array_length - 1:
                slice_index2 += 1
            else:
                slice_index1 -= 1

        return slice_index1, slice_index2

    def pmx(self, parent1, parent2, slice_index1, slice_index2):
        if len(parent1) != len(parent2):
            raise ValueError('Parents have not the same length!')

        child1, child2 = parent1.copy(), parent2.copy()
        mapping1, mapping2 = {}, {}

        # Map Slice
        for i in range(slice_index1, slice_index2):
            mapping1[parent2[i]] = parent1[i]
            child1[i] = parent2[i]

            mapping2[parent1[i]] = parent2[i]
            child2[i] = parent1[i]

        # Repair Lower Slice
        for i in range(slice_index1):
            repair_count = 0
            while child1[i] in mapping1:
                if mapping1[child1[i]] == child1[i] or repair_count > len(parent1):
                    break
                child1[i] = mapping1[child1[i]]
                repair_count += 1

            repair_count = 0
            while child2[i] in mapping2:
                if mapping2[child2[i]] == child2[i] or repair_count > len(parent1):
                    break
                child2[i] = mapping2[child2[i]]
                repair_count += 1

        # Repair Upper Slice
        for i in range(slice_index2, len(child1)):
            repair_count = 0
            while child1[i] in mapping1:
                if mapping1[child1[i]] == child1[i] or repair_count > len(parent1):
                    break
                child1[i] = mapping1[child1[i]]
                repair_count += 1

            repair_count = 0
            while child2[i] in mapping2:
                if mapping2[child2[i]] == child2[i] or repair_count > len(parent1):
                    break
                child2[i] = mapping2[child2[i]]
                repair_count += 1

        return child1, child2

    def pmx_crossover_function(self, parents, size, ga_instance):
        offspring = np.empty(size)
        num_offspring, num_genes = size

        for k in range(num_offspring):
            parent1 = parents[k % parents.shape[0], :]
            parent2 = parents[(k + 1) % parents.shape[0], :]

            # Get random slice indices
            slice_index1, slice_index2 = self.get_random_slice_indices(
                num_genes)

            # Perform PMX Crossover
            child1, child2 = self.pmx(
                parent1, parent2, slice_index1, slice_index2)

            offspring[k, :] = child1
            if k+1 < num_offspring:
                offspring[k+1, :] = child2
                k += 1  # Skip next iteration as we already filled the next offspring

        return offspring

    def run(self):
        initial_population = self.create_initial_population()
        average_distance_initial_population = self.calculate_average_total_distance(
            initial_population)
        print("Average Total Distance of Initial Population:",
              average_distance_initial_population)

        ga_instance = pygad.GA(
            num_generations=20000,
            num_parents_mating=2,
            fitness_func=self.fitness_func,  # Pass the method bound to the current instance
            sol_per_pop=100,
            num_genes=self.num_customers,
            gene_space=[i for i in range(self.num_customers)],
            mutation_percent_genes=10,
            parent_selection_type="tournament",
            mutation_type="random",
            on_generation=self.on_generation,
            crossover_type=self.pmx_crossover_function,
            keep_parents=2,
            crossover_probability=0.8,
            mutation_probability=0.1,
            # stop_criteria=["saturate_100"],
            initial_population=initial_population
        )

        ga_instance.run()
        solution, solution_fitness, _ = ga_instance.best_solution()
        print("Best Solution (Route):", solution)
        print("Best Solution Fitness:", solution_fitness)
        if (self.are_all_customers_visited(solution)):
            print(f"All customers are visited!\n")
        else:
            print(f"FAILED")
        best_routes = self.generate_routes(solution)
        print(f"Best Routes: {best_routes}")
        total_distance = self.calculate_total_distance(best_routes)
        print(f"Total Distance: {total_distance}\n\n")
        self.best_solution_cost = total_distance
        # self.draw_graph(best_routes) tmp

    def draw_graph(self, routes):
        G = nx.Graph()
        for route in routes:
            # Append depot index at the start and end of each route
            route_with_depot = [self.depot_index] + route + [self.depot_index]
            for i in range(len(route_with_depot) - 1):
                # Draw edge between consecutive nodes in the route
                G.add_edge(route_with_depot[i], route_with_depot[i + 1])

        pos = nx.spring_layout(G)  # positions for all nodes
        nx.draw_networkx_nodes(G, pos, node_size=700)
        nx.draw_networkx_edges(G, pos, edgelist=G.edges(), width=2)
        nx.draw_networkx_labels(G, pos, font_size=20, font_family="sans-serif")

        plt.axis("off")
        plt.show()


class VRPProblem:
    def __init__(self, file_path):
        self.file_path = file_path
        self.depot_index = None
        self.num_vehicles = None
        self.capacity = None
        self.optimal_value = None  # Add this line to initialize the optimal value
        self.node_coord_section = []
        self.demand_section = []
        self.parse_problem()

    def parse_problem(self):
        with open(self.file_path, 'r') as file:
            lines = file.read().splitlines()

        for line in lines:
            if line.startswith("COMMENT"):
                parts = line.split(',')
                self.num_vehicles = int(parts[1].split(':')[1].strip())
                self.optimal_value = float(parts[2].split(
                    ':')[1].strip().replace(")", ""))  # Extract the optimal value
                break

        capacity_start = next((i for i, element in enumerate(
            lines) if element.startswith("CAPACITY")), None)
        coord_start = lines.index("NODE_COORD_SECTION ") + 1
        demand_start = lines.index("DEMAND_SECTION ") + 1
        depot_section_start = lines.index("DEPOT_SECTION ") + 1

        self.capacity = lines[capacity_start].split(":")[1].strip()

        self.depot_index = lines[depot_section_start].strip()
        for line in lines[coord_start:demand_start - 1]:
            parts = line.split()
            self.node_coord_section.append(
                (int(parts[0]), float(parts[1]), float(parts[2])))

        for line in lines[demand_start:depot_section_start - 1]:
            self.demand_section.append((int(parts[0]), int(parts[1])))
            parts = line.split()


class TestRunner:
    def __init__(self, folder_path):
        self.folder_path = folder_path
        self.results = []
        self.first_test_stats = {
            "min_cost": None,
            "max_cost": None,
            "avg_cost": None,
            "times": []
        }

        self.second_test_stats = {
            "min_cost": None,
            "max_cost": None,
            "avg_cost": None,
            "times": []
        }

    def run_tests(self):
        problem_files = [f for f in os.listdir(
            self.folder_path) if f.endswith('.vrp')]

        for test_idx, problem_file in enumerate(problem_files, start=1):
            problem_path = os.path.join(self.folder_path, problem_file)
            problem = VRPProblem(problem_path)
            # print("----------------------------------------------------------")
            # print(
            #     f"********************* TEST {test_idx} *************************")
            costs = []
            times = []

            # if (test_idx == 1):
            #     for i in range(10):
            #         cvrp_test = CVRPTest(problem)
            #         start_time = time.time()
            #         cvrp_test.run()
            #         end_time = time.time()
            #         costs.append(cvrp_test.best_solution_cost)
            #         times.append(end_time - start_time)

            #     self.first_test_stats["min_cost"] = min(costs)
            #     self.first_test_stats["max_cost"] = max(costs)
            #     self.first_test_stats["avg_cost"] = sum(costs) / len(costs)
            #     self.first_test_stats["times"] = times
            num_runs = 1
            if (test_idx == 2):
                for i in range(num_runs):
                    cvrp_test = CVRPTest(problem)
                    start_time = time.time()
                    cvrp_test.run()
                    end_time = time.time()
                    costs.append(cvrp_test.best_solution_cost)
                    times.append(end_time - start_time)
              #  cvrp_test.plot_fitness_over_generations()
                cvrp_test.plot_distance_over_generations()
                self.second_test_stats["min_cost"] = min(costs)
                self.second_test_stats["max_cost"] = max(costs)
                self.second_test_stats["avg_cost"] = sum(costs) / len(costs)
                self.second_test_stats["times"] = times
                difference = cvrp_test.best_solution_cost - problem.optimal_value
                self.results.append({
                    "Problem": test_idx,
                    "Optimal Value": problem.optimal_value,
                    "GA Solution Cost": cvrp_test.best_solution_cost,
                    "Difference": difference,
                    "Time Taken": times[-1],
                    "Status": "Optimal" if difference == 0 else "Suboptimal"
                })
            # else:
            #     break
            #     cvrp_test = CVRPTest(problem)
            #     start_time = time.time()
            #     cvrp_test.run()
            #     end_time = time.time()
            #     costs.append(cvrp_test.best_solution_cost)
            #     times.append(end_time - start_time)

            # Calculate the difference between the GA solution and the optimal value
            # difference = cvrp_test.best_solution_cost - problem.optimal_value
            # self.results.append({
            #     "Problem": test_idx,
            #     "Optimal Value": problem.optimal_value,
            #     "GA Solution Cost": cvrp_test.best_solution_cost,
            #     "Difference": difference,
            #     "Time Taken": times[-1],
            #     "Status": "Optimal" if difference == 0 else "Suboptimal"
            # })

    def plot_results(self):
        # Extract problem names, GA solution costs, optimal costs, and times from the results
        problems = [result["Problem"] for result in self.results]
        ga_costs = [result["GA Solution Cost"] for result in self.results]
        optimal_costs = [result["Optimal Value"] for result in self.results]
        times = [result["Time Taken"] for result in self.results]

        # Create a figure and a set of subplots
        fig, ax1 = plt.subplots()

        # Set the positions of the bars
        width = 0.35  # Width of the bars
        indices = np.arange(len(problems))  # The label locations

        # Plotting the GA solution costs and optimal costs
        rects1 = ax1.bar(indices - width/2, ga_costs, width,
                         label='GA Solution Cost', color='tab:red')
        rects2 = ax1.bar(indices + width/2, optimal_costs,
                         width, label='Optimal Cost', color='tab:blue')

        ax1.set_xlabel('Problem')
        ax1.set_ylabel('Cost')
        ax1.set_xticks(indices)
        ax1.set_xticklabels(problems)
        ax1.legend()

        # Instantiate a second y-axis sharing the same x-axis
        time_x_positions = indices
        ax2 = ax1.twinx()
        color = 'tab:green'
        ax2.set_ylabel('Time Taken (seconds)', color=color)
        ax2.plot(time_x_positions, times, color=color, marker='o',
                 linestyle='dashed', linewidth=2)
        ax2.tick_params(axis='y', labelcolor=color)

        buffer = 5
        ax2.set_ylim([min(times) - buffer, max(times) + buffer])

        # Show plot
        fig.tight_layout()
        plt.xticks(rotation=45)  # Rotate problem names for better readability
        plt.show()

    def plot_first_test_stats(self):
        stats = self.first_test_stats
        fig, ax1 = plt.subplots()

        # Cost values: min, max, average, and optimal
        costs = [stats["min_cost"], stats["max_cost"],
                 stats["avg_cost"], self.results[0]["Optimal Value"]]
        cost_labels = ['Min Cost', 'Max Cost', 'Avg Cost', 'Optimal Cost']

        # Bar plot for min, max, average, and optimal costs
        ax1.bar(cost_labels, costs, color='tab:blue')

        ax1.set_ylabel('Cost', color='tab:blue')
        ax1.tick_params(axis='y', labelcolor='tab:blue')
        ax1.set_title('First Test Statistics')

        plt.show()

    def plot_second_test_stats(self):
        stats = self.second_test_stats
        fig, ax1 = plt.subplots()

        # Cost values: min, max, average, and optimal
        costs = [stats["min_cost"], stats["max_cost"],
                 stats["avg_cost"], self.results[0]["Optimal Value"]]
        cost_labels = ['Min Cost', 'Max Cost', 'Avg Cost', 'Optimal Cost']

        # Bar plot for min, max, average, and optimal costs
        ax1.bar(cost_labels, costs, color='tab:blue')

        ax1.set_ylabel('Cost', color='tab:blue')
        ax1.tick_params(axis='y', labelcolor='tab:blue')
        ax1.set_title('Third Test Statistics')

        plt.show()


# Usage
test_runner = TestRunner('./input/Vrp-Set-A/A/')
test_runner.run_tests()
# test_runner.plot_results()
# test_runner.plot_first_test_stats()
# test_runner.plot_second_test_stats()
